@@ -101,9 +101,13 @@ class DataBunch(object):
         # check X_train, y_train, X_test
         if self.check_data_format(X_train):
             self.X_train_source = pd.DataFrame(X_train)
+            self.X_train_source = remove_duplicate_cols_in_dataset(self.X_train_source)
         if X_test is not None:
             if self.check_data_format(X_test):
                 self.X_test_source = pd.DataFrame(X_test)
+                self.X_test_source = remove_duplicate_cols_in_dataset(self.X_test_source)
+
+
         ### There is a chance for an error in this - so worth watching!
         if y_train is not None:
             le = LabelEncoder()
@@ -208,7 +212,7 @@ class DataBunch(object):
             nan_columns = list(data[cols].columns[data[cols].isnull().sum() > 0])
             if nan_columns:
                 for nan_column in nan_columns:
-                    data[nan_column+'isNAN'] = pd.isna(data[nan_column]).astype('uint8')
+                    data[nan_column+'_isNAN'] = pd.isna(data[nan_column]).astype('uint8')
                 data.fillna(data.median(), inplace=True)
         return(data)
 
@@ -397,6 +401,8 @@ class DataBunch(object):
             X_test (pd.DataFrame)
 
         """
+        #### Sometimes there are duplicates in column names. You must remove them here. ###
+        cat_features = find_remove_duplicates(cat_features)
 
         # concat datasets for correct processing.
         df_train = X_train.copy()
@@ -406,8 +412,11 @@ class DataBunch(object):
             test_data = None ### Set test_data to None if X_test is None
         else:
             test_data = X_test.copy()
+            test_data = remove_duplicate_cols_in_dataset(test_data)
             data = copy.deepcopy(df_train)
 
+        data = remove_duplicate_cols_in_dataset(data)
+
         # object & num features
         object_features = list(data.columns[(data.dtypes == 'object') | (data.dtypes == 'category')])
         num_features = list(set(data.columns) - set(cat_features) - set(object_features) - {'test'})
@@ -421,7 +430,9 @@ class DataBunch(object):
         self.binary_features_names = []
 
         # LabelEncode all Binary Features - leave the rest alone
-        for feature in data.columns:
+        cols = data.columns.tolist()
+        #### This sometimes errors because there are duplicate columns in a dataset ###
+        for feature in cols:
             if (feature != 'test') and (data[feature].nunique(dropna=False) < 3):
                 data[feature] = data[feature].astype('category').cat.codes
                 if test_data is not None:
@@ -429,36 +440,39 @@ class DataBunch(object):
                 self.binary_features_names.append(feature)
 
         # Convert all Category features "Category" type variables if no encoding is specified
-        if encodet_features_names:
-            if cat_encoder_names is None:
-                for feature in encodet_features_names:
-                    data[feature] = data[feature].astype('category').cat.codes
-                    if test_data is not None:
-                        test_data[feature] = test_data[feature].astype('category').cat.codes
-            else:
-                #### If an encoder is specified, then use that encoder to transform categorical variables
-                if verbose > 0:
-                    print('> Generate Categorical Encoded features')
-
-                copy_cat_encoder_names = copy.deepcopy(cat_encoder_names)
-                for encoder_name in copy_cat_encoder_names:
+        cat_only_encoders = [x for x in self.cat_encoder_names if x in self.cat_encoders_names]
+        if len(cat_only_encoders) > 0:
+            ### Just skip if this encoder is not in the list of category encoders ##
+            if encodet_features_names:
+                if cat_encoder_names is None:
+                    for feature in encodet_features_names:
+                        data[feature] = data[feature].astype('category').cat.codes
+                        if test_data is not None:
+                            test_data[feature] = test_data[feature].astype('category').cat.codes
+                else:
+                    #### If an encoder is specified, then use that encoder to transform categorical variables
                     if verbose > 0:
-                        print(' + To know more, click: %s' %self.cat_encoders_names[encoder_name][1])
-                    data_encodet, train_encoder = self.gen_cat_encodet_features(data[encodet_features_names],
-                                                                encoder_name)
-                    if not isinstance(data_encodet, str):
-                        data = data.join(data_encodet)
-                    if test_data is not None:
-                        test_encodet, _ = self.gen_cat_encodet_features(test_data[encodet_features_names],
-                                                                train_encoder)
-                        if not isinstance(test_encodet, str):
-                            test_data = test_data.join(test_encodet)
+                        print('> Generate Categorical Encoded features')
 
-                    if verbose > 0:
+                    copy_cat_encoder_names = copy.deepcopy(cat_encoder_names)
+                    for encoder_name in copy_cat_encoder_names:
+                        if verbose > 0:
+                            print(' + To know more, click: %s' %self.cat_encoders_names[encoder_name][1])
+                        data_encodet, train_encoder = self.gen_cat_encodet_features(data[encodet_features_names],
+                                                                    encoder_name)
                         if not isinstance(data_encodet, str):
-                            addl_features = data_encodet.shape[1] - original_number_features
-                            count_number_features += addl_features
-                            print(' + added ', addl_features, ' additional Features using',encoder_name)
+                            data = data.join(data_encodet)
+                        if test_data is not None:
+                            test_encodet, _ = self.gen_cat_encodet_features(test_data[encodet_features_names],
+                                                                    train_encoder)
+                            if not isinstance(test_encodet, str):
+                                test_data = test_data.join(test_encodet)
+
+                        if verbose > 0:
+                            if not isinstance(data_encodet, str):
+                                addl_features = data_encodet.shape[1] - original_number_features
+                                count_number_features += addl_features
+                                print(' + added ', addl_features, ' additional Features using',encoder_name)
 
         # Generate Target related Encoder features for cat variables:
         target_encoders = [x for x in self.cat_encoder_names if x in self.target_encoders_names_list]
@@ -568,13 +582,13 @@ class DataBunch(object):
             ### If they have been transformed into Cat Encoded variables, then you can drop them!
             data.drop(columns=encodet_features_names, inplace=True)
         data.replace([np.inf, -np.inf], np.nan, inplace=True)
-        data.fillna(0, inplace=True)
+        #data.fillna(0, inplace=True)
         if test_data is not None:
             if not len(cat_encoder_names) == 0:
                 ### if there is no categorical encoding, then let the categorical_vars pass through.
                 test_data.drop(columns=encodet_features_names, inplace=True)
             test_data.replace([np.inf, -np.inf], np.nan, inplace=True)
-            test_data.fillna(0, inplace=True)
+            #test_data.fillna(0, inplace=True)
 
         X_train = copy.deepcopy(data)
         X_test = copy.deepcopy(test_data)
@@ -595,6 +609,9 @@ class DataBunch(object):
             print('> Final Number of Features: ', (X_train.shape[1]))
             print('#'*50)
             print('New X_train shape: ', X_train.shape, '| X_test shape: ', X_test.shape)
+            if len(left_subtract(X_test.columns, X_train.columns)) > 0:
+                print("""There are more columns in test than train 
+                    due to missing columns being more in test than train. Continuing...""")
 
         return (X_train, X_test)
 ################################################################################
@@ -614,3 +631,22 @@ def left_subtract(l1,l2):
             lst.append(i)
     return lst
 #################################################################################
+def remove_duplicate_cols_in_dataset(df):
+    df = copy.deepcopy(df)
+    cols = df.columns.tolist()
+    number_duplicates = df.columns.duplicated().astype(int).sum()
+    if  number_duplicates > 0:
+        print('Detected %d duplicate columns in dataset. Removing duplicates...' %number_duplicates)
+        df = df.loc[:,~df.columns.duplicated()]
+    return df
+###########################################################################
+# Removes duplicates from a list to return unique values - USED ONLYONCE
+def find_remove_duplicates(values):
+    output = []
+    seen = set()
+    for value in values:
+        if value not in seen:
+            output.append(value)
+            seen.add(value)
+    return output
+#################################################################################
